(******************************************************************************
 * Citadelle
 *
 * Copyright (c) 2011-2019 Université Paris-Saclay, Univ. Paris-Sud, France
 *               2013-2017 IRT SystemX, France
 *               2011-2015 Achim D. Brucker, Germany
 *               2016-2018 The University of Sheffield, UK
 *               2016-2017 Nanyang Technological University, Singapore
 *               2017-2018 Virginia Tech, USA
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)

chapter \<open>Appendix: Citadelle's Meta Interface\<close>

theory Generator_dynamic_sequential
  imports Core
          Isabelle_Meta_Model.Isabelle_code_runtime
          Isabelle_Meta_Model.Isabelle_typedecl
  keywords (* Language (added) *)
           "self"
           "with_only"
      
           (* Isabelle syntax *)
           "output_directory"
           "THEORY" "IMPORTS" "SECTION" "SORRY" "no_dirty"
           "deep" "shallow" "syntax_print" "skip_export"
           "generation_semantics"
           "flush_all"
      
           (* Isabelle semantics (parameterizing the semantics of Language) *)
           "design" "analysis" "oid_start"

       and (* Language *)
           "meta_command" "meta_command'"

           (* Isabelle syntax *)
           "generation_syntax"

           :: thy_decl
begin
\<comment> \<open>Derived from: \<^file>\<open>../../../../Citadelle/src/compiler/Generator_dynamic_sequential.thy\<close>\<close>

section \<open>\<close>

ML \<open>
structure META = C_Ast
\<close>

subsection\<open>Interface Between the Reflected and the Native\<close>

ML\<open>
val To_string0 = META.meta_of_logic
val To_nat = Code_Numeral.integer_of_natural

exception THY_REQUIRED of Position.T
fun get_thy pos f = fn NONE => raise (THY_REQUIRED pos) | SOME thy => f thy

infix 1 #~> |>::
fun f #~> g = uncurry g oo f
fun x |>:: f = cons f x
\<close>

ML\<open>
structure From = struct
 val string = META.SS_base o META.ST
 val binding = string o Binding.name_of
 (*fun term ctxt s = string (YXML.content_of (Syntax.string_of_term ctxt s))*)
 val nat = Code_Numeral.natural_of_integer
 val internal_oid = META.Oid o nat
 val option = Option.map
 val list = List.map
 fun pair f1 f2 (x, y) = (f1 x, f2 y)
 fun pair3 f1 f2 f3 (x, y, z) = (f1 x, f2 y, f3 z)

 structure Pure = struct
 val indexname = pair string nat
 val class = string
 val sort = list class
 fun typ e = (fn
     Type (s, l) => (META.Typeb o pair string (list typ)) (s, l)
   | TFree (s, s0) => (META.TFree o pair string sort) (s, s0)
   | TVar (i, s0) => (META.TVara o pair indexname sort) (i, s0)
  ) e
 fun term e = (fn
     Const (s, t) => (META.Consta o pair string typ) (s, t)
   | Free (s, t) => (META.Free o pair string typ) (s, t)
   | Var (i, t) => (META.Var o pair indexname typ) (i, t)
   | Bound i => (META.Bound o nat) i
   | Abs (s, ty, t) => (META.Absa o pair3 string typ term) (s, ty, t)
   | op $ (term1, term2) => (META.Appa o pair term term) (term1, term2)
  ) e
 end

 fun read_term thy expr =
   META.T_pure (Pure.term (Syntax.read_term (get_thy \<^here> Proof_Context.init_global thy) expr), SOME (string expr))
end
\<close>

ML\<open>
fun List_mapi f = META.mapi (f o To_nat)
fun out_intensify s1 s2 = Output.state ((s1 |> Markup.markup Markup.intensify) ^ s2)
fun out_intensify' tps fmt = out_intensify (Timing.message (Timing.result tps) |> Markup.markup fmt)

structure Toplevel' = struct
  datatype state_read = Load_backup | Load_previous
  datatype state_write = Store_backup | Store_default

  datatype toplevel = Theory of theory -> theory
                    | Keep of theory -> unit
                    | Read_Write of state_read * state_write

  structure T = struct
    val theory = cons o Theory
    val keep = cons o Keep
    val read_write = cons o Read_Write
  end

  val keep_theory = T.keep
  fun keep f tr = (\<^command_keyword>\<open>print_syntax\<close>, T.keep f) :: tr
  fun read_write_keep rw = (\<^command_keyword>\<open>setup\<close>, fn tr => tr |> T.read_write rw |> T.keep (K ()))
  fun setup_theory (res, tr) f = rev ((\<^command_keyword>\<open>setup\<close>, T.theory (f res)) :: tr)
  fun keep_output tps fmt msg = cons (\<^command_keyword>\<open>print_syntax\<close>, T.keep (fn _ => out_intensify' tps fmt msg))
end

structure Outer_Syntax' = struct
  fun command' f thy =
    fold snd (f thy NONE) [] |> rev
                             |> (fn tr => fold (fn Toplevel'.Theory f => f
                                                 | Toplevel'.Keep f => tap f
                                                 | Toplevel'.Read_Write _ => I) tr thy)
  fun command name_pos comment parse =
    Outer_Syntax.command name_pos comment (parse >> (Toplevel.theory o command'))
end

structure Resources' = struct
  fun formal_check check_file ctxt opt_dir source =
    let
      val name = Input.string_of source;
      val pos = Input.pos_of source;
      val delimited = Input.is_delimited source;
  
      val _ = Context_Position.report ctxt pos (Markup.language_path delimited);
  
      fun err msg = error (msg ^ Position.here pos);
      val dir =
        (case opt_dir of
          SOME dir => dir
        | NONE => Resources.master_directory (Proof_Context.theory_of ctxt));
      val path = dir + Path.explode name handle ERROR msg => err msg;
      val path = Path.expand path handle ERROR msg => err msg;
      val path' = Path.implode path;
      val _ = Context_Position.report ctxt pos (Markup.path path');
      val _ : Path.T = check_file path handle ERROR msg => err msg;
    in path' end;
  
  val check_file = formal_check File.check_file;
  val check_dir = formal_check File.check_dir;
end
\<close>

ML\<open>
structure Ty' = struct
fun check l_oid l =
  let val Mp = META.map_prod
      val Me = String.explode
      val Mi = String.implode
      val Ml = map in
  META.check_export_code
    (writeln o Mi)
    (warning o Mi)
    (fn s => writeln (Markup.markup (Markup.bad ()) (Mi s)))
    (error o To_string0)
    (Ml (Mp I Me) l_oid)
    ((META.SS_base o META.ST) l)
  end
end
\<close>

subsection\<open>Binding of the Reflected API to the Native API\<close>

ML\<open>
structure META_overload = struct
  val of_semi__typ = META.of_semi_typ To_string0
  val of_semi__term = META.of_semi_terma To_string0
  val of_semi__term' = META.of_semi_term To_string0
  val fold = fold
end
\<close>

ML\<open>
type ('a, 'b) toplevel_dual = { par: 'a, seq: 'b }
type ('transitionM, 'Proof_stateM, 'state) toplevel =
  { context_of: 'state -> local_theory

  , keep: ('state -> unit) -> 'transitionM
  , generic_theory: (generic_theory -> generic_theory) -> 'transitionM
  , theory: (theory -> theory) -> 'transitionM
  , begin_main_target: bool -> (theory -> local_theory) -> 'transitionM
  , local_theory': (bool * Position.T) option -> (xstring * Position.T) option ->
      (bool -> local_theory -> local_theory) -> 'transitionM
  , local_theory: (bool * Position.T) option -> (xstring * Position.T) option ->
      (local_theory -> local_theory) -> 'transitionM
  , local_theory_to_proof': (bool * Position.T) option -> (xstring * Position.T) option ->
      (bool -> local_theory -> Proof.state) -> 'transitionM
  , local_theory_to_proof: (bool * Position.T) option -> (xstring * Position.T) option ->
      (local_theory -> Proof.state) -> 'transitionM
  , proof': (bool -> Proof.state -> Proof.state) -> 'Proof_stateM
  , proofs: (Proof.state -> Proof.state Seq.result Seq.seq) -> 'Proof_stateM
  , proof: (Proof.state -> Proof.state) -> 'Proof_stateM
  (* *)
  , tr_report: Method.text_range -> 'transitionM -> 'transitionM
  , tr_report_o: Method.text_range option -> 'transitionM -> 'transitionM
  , tr_raw: (Toplevel.transition -> Toplevel.transition) -> 'transitionM
  , pr_report: Method.text_range -> 'Proof_stateM -> 'Proof_stateM
  , pr_report_o: Method.text_range option -> 'Proof_stateM -> 'Proof_stateM
  , dual: (Toplevel.transition -> Toplevel.transition, Proof.state -> Proof.state) toplevel_dual -> 'Proof_stateM }

structure Bind_Isabelle = struct
fun To_binding s = Binding.make (s, Position.none)
val To_sbinding = To_binding o To_string0

fun semi__method_simp g f = Method.Basic (fn ctxt => SIMPLE_METHOD (g (asm_full_simp_tac (f ctxt))))
val semi__method_simp_one = semi__method_simp (fn f => f 1)
val semi__method_simp_all = semi__method_simp (CHANGED_PROP o PARALLEL_GOALS o ALLGOALS)

datatype semi__thm' = Thms_single' of thm
                    | Thms_mult' of thm list

fun semi__thm_attribute ctxt = let open META open META_overload val S = fn Thms_single' t => t in
 fn Thm_thm s => Thms_single' (Proof_Context.get_thm ctxt (To_string0 s))
  | Thm_thms s => Thms_mult' (Proof_Context.get_thms ctxt (To_string0 s))
  | Thm_THEN (e1, e2) =>
      (case (semi__thm_attribute ctxt e1, semi__thm_attribute ctxt e2) of
         (Thms_single' e1, Thms_single' e2) => Thms_single' (e1 RSN (1, e2))
       | (Thms_mult' e1, Thms_mult' e2) => Thms_mult' (e1 RLN (1, e2)))
  | Thm_simplified (e1, e2) =>
      Thms_single' (asm_full_simplify (clear_simpset ctxt addsimps [S (semi__thm_attribute ctxt e2)])
                                      (S (semi__thm_attribute ctxt e1)))
  | Thm_OF (e1, e2) =>
      Thms_single' ([S (semi__thm_attribute ctxt e2)] MRS (S (semi__thm_attribute ctxt e1)))
  | Thm_where (nth, l) =>
      Thms_single' (Rule_Insts.where_rule
                      ctxt
                      (List.map (fn (var, expr) =>
                                   (((To_string0 var, 0), Position.none), of_semi__term expr)) l)
                      []
                      (S (semi__thm_attribute ctxt nth)))
  | Thm_symmetric e1 =>
      let val e2 = S (semi__thm_attribute ctxt (Thm_thm (From.string "sym"))) in
        case semi__thm_attribute ctxt e1 of
          Thms_single' e1 => Thms_single' (e1 RSN (1, e2))
        | Thms_mult' e1 => Thms_mult' (e1 RLN (1, [e2]))
      end
  | Thm_of (nth, l) =>
      Thms_single' (Rule_Insts.of_rule
                     ctxt
                     (List.map (SOME o of_semi__term) l, [])
                     []
                     (S (semi__thm_attribute ctxt nth)))
end

fun semi__thm_attribute_single ctxt s = case (semi__thm_attribute ctxt s) of Thms_single' t => t

fun semi__thm_mult ctxt =
  let fun f thy = case (semi__thm_attribute ctxt thy) of Thms_mult' t => t
                                                  | Thms_single' t => [t] in
  fn META.Thms_single thy => f thy
   | META.Thms_mult thy => f thy
  end

fun semi__thm_mult_l ctxt l = List.concat (map (semi__thm_mult ctxt) l)

fun semi__method_simp_only l ctxt = clear_simpset ctxt addsimps (semi__thm_mult_l ctxt l)
fun semi__method_simp_add_del_split (l_add, l_del, l_split) ctxt =
  fold Splitter.add_split (semi__thm_mult_l ctxt l_split)
                          (ctxt addsimps (semi__thm_mult_l ctxt l_add)
                                delsimps (semi__thm_mult_l ctxt l_del))

fun semi__method expr = let open META open Method open META_overload in case expr of
    Method_rule o_s => Basic (fn ctxt =>
      METHOD (HEADGOAL o Classical.rule_tac
                           ctxt
                           (case o_s of NONE => []
                                      | SOME s => [semi__thm_attribute_single ctxt s])))
  | Method_drule s => Basic (fn ctxt => drule ctxt 0 [semi__thm_attribute_single ctxt s])
  | Method_erule s => Basic (fn ctxt => erule ctxt 0 [semi__thm_attribute_single ctxt s])
  | Method_elim s => Basic (fn ctxt => elim ctxt [semi__thm_attribute_single ctxt s])
  | Method_intro l => Basic (fn ctxt => intro ctxt (map (semi__thm_attribute_single ctxt) l))
  | Method_subst (asm, l, s) => Basic (fn ctxt =>
      SIMPLE_METHOD' ((if asm then EqSubst.eqsubst_asm_tac else EqSubst.eqsubst_tac)
                        ctxt
                        (map (the o Int.fromString o To_string0) l)
                        [semi__thm_attribute_single ctxt s]))
  | Method_insert l => Basic (fn ctxt => insert (semi__thm_mult_l ctxt l))
  | Method_plus t => Combinator ( no_combinator_info
                                , Repeat1
                                , [Combinator (no_combinator_info, Then, List.map semi__method t)])
  | Method_option t => Combinator ( no_combinator_info
                                  , Try
                                  , [Combinator (no_combinator_info, Then, List.map semi__method t)])
  | Method_or t => Combinator (no_combinator_info, Orelse, List.map semi__method t)
  | Method_one (Method_simp_only l) => semi__method_simp_one (semi__method_simp_only l)
  | Method_one (Method_simp_add_del_split l) => semi__method_simp_one (semi__method_simp_add_del_split l)
  | Method_all (Method_simp_only l) => semi__method_simp_all (semi__method_simp_only l)
  | Method_all (Method_simp_add_del_split l) => semi__method_simp_all (semi__method_simp_add_del_split l)
  | Method_auto_simp_add_split (l_simp, l_split) =>
      Basic (fn ctxt => SIMPLE_METHOD (auto_tac (fold (fn (f, l) => fold f l)
              [(Simplifier.add_simp, semi__thm_mult_l ctxt l_simp)
              ,(Splitter.add_split, List.map (Proof_Context.get_thm ctxt o To_string0) l_split)]
              ctxt)))
  | Method_rename_tac l => Basic (K (SIMPLE_METHOD' (Tactic.rename_tac (List.map To_string0 l))))
  | Method_case_tac e =>
      Basic (fn ctxt => SIMPLE_METHOD' (Induct_Tacs.case_tac ctxt (of_semi__term e) [] NONE))
  | Method_blast n =>
      Basic (case n of NONE => SIMPLE_METHOD' o blast_tac
                     | SOME lim => fn ctxt => SIMPLE_METHOD' (depth_tac ctxt (To_nat lim)))
  | Method_clarify => Basic (fn ctxt => (SIMPLE_METHOD' (fn i => CHANGED_PROP (clarify_tac ctxt i))))
  | Method_metis (l_opt, l) =>
      Basic (fn ctxt => (METHOD oo Metis_Tactic.metis_method)
                          ( (if l_opt = [] then NONE else SOME (map To_string0 l_opt), NONE)
                          , map (semi__thm_attribute_single ctxt) l)
                          ctxt)
end

fun then_tactic l = let open Method in
  (Combinator (no_combinator_info, Then, map semi__method l), (Position.none, Position.none))
end

fun terminal_proof0 f1 f2 f3 top o_by = let open META in case o_by of
   Command_done =>       (\<^command_keyword>\<open>done\<close>, #dual top { par = Isar_Cmd.done_proof
                                                            , seq = f1 })
 | Command_sorry =>      (\<^command_keyword>\<open>sorry\<close>, #dual top { par = Isar_Cmd.skip_proof
                                                             , seq = f2 true })
 | Command_by l_apply => (\<^command_keyword>\<open>by\<close>, let val (m1, m2) = (then_tactic l_apply, NONE) in
                                                #pr_report top m1
                                                  (#pr_report_o top m2
                                                    (#dual top { par = Isar_Cmd.terminal_proof (m1, m2)
                                                               , seq = f3 (m1, m2) })) end)
end

fun terminal_proof_dual top =
  terminal_proof0 Proof.local_done_proof Proof.local_skip_proof Proof.local_terminal_proof top

fun proof_show_gen top f (thes, thes_when) st = st
  |>:: (\<^command_keyword>\<open>proof\<close>,
      let val m = SOME ( Method.Source [Token.make_string ("-", Position.none)]
                       , (Position.none, Position.none)) in
      (#pr_report_o top m (#proofs top (Proof.proof m))) end)
  |> f
  |>:: (\<^command_keyword>\<open>show\<close>, #proof' top (fn int => Proof.show_cmd
       (thes_when = [])
       NONE
       (K I)
       []
       (if thes_when = [] then [] else [(Binding.empty_atts, map (fn t => (t, [])) thes_when)])
       [(Binding.empty_atts, [(thes, [])])]
       int #> #2))

fun semi__command_state top (META.Command_apply_end l) = let open META_overload in
  cons (\<^command_keyword>\<open>apply_end\<close>, let val m = then_tactic l in
    #pr_report top m (#proofs top (Proof.apply_end m)) end)
end

fun semi__command_proof top = let open META_overload
  val thesis = "?thesis"
  fun cons_proof_show f = proof_show_gen top f (thesis, [])
  fun cons_let (e1, e2) =
        cons (\<^command_keyword>\<open>let\<close>, #proof top
          (Proof.let_bind_cmd [([of_semi__term e1], of_semi__term e2)])) in
  fn META.Command_apply l =>
        cons (\<^command_keyword>\<open>apply\<close>, let val m = then_tactic l in
          #pr_report top m (#proofs top (Proof.apply m)) end)
   | META.Command_using l =>
        cons (\<^command_keyword>\<open>using\<close>, #proof top (fn st =>
          Proof.using [map (fn s => ([s], [])) (semi__thm_mult_l (Proof.context_of st) l)] st))
   | META.Command_unfolding l =>
        cons (\<^command_keyword>\<open>unfolding\<close>, #proof top (fn st =>
          Proof.unfolding [map (fn s => ([s], [])) (semi__thm_mult_l (Proof.context_of st) l)] st))
   | META.Command_let e =>
        cons_proof_show (cons_let e)
   | META.Command_have (n, b, e, e_pr) => (fn st => st
     |> cons_proof_show (fn st => st
       |>:: (\<^command_keyword>\<open>have\<close>, #proof' top (fn int =>
          Proof.have_cmd true NONE (K I) [] []
            [( (To_sbinding n, if b then [[Token.make_string ("simp", Position.none)]] else [])
             , [(of_semi__term e, [])])] int #> #2))
       |>:: terminal_proof_dual top e_pr))
   | META.Command_fix_let (l, l_let, o_exp, _) => (fn st => st
     |> proof_show_gen top (fn st => st
       |>:: (\<^command_keyword>\<open>fix\<close>, #proof top
            (Proof.fix_cmd (List.map (fn i => (To_sbinding i, NONE, NoSyn)) l)))
       |> fold cons_let l_let)
          ( case o_exp of NONE => thesis | SOME (l_spec, _) =>
             (String.concatWith (" \<Longrightarrow> ")
                                (List.map of_semi__term l_spec))
          , case o_exp of NONE => [] | SOME (_, l_when) => List.map of_semi__term l_when))
end

fun end' top =
 (\<^command_keyword>\<open>end\<close>, #tr_raw top (Toplevel.exit o Toplevel.end_main_target o Toplevel.end_nested_target o
                                      Toplevel.end_proof (K Proof.end_notepad)))

structure Cmd = struct open META open META_overload
fun input_source ml = Input.source false (of_semi__term' ml) (Position.none, Position.none)

fun datatype' top (Datatypea (version, l)) = \<^cancel>\<open>case version of Datatype_new =>\<close> #local_theory top NONE NONE
  (BNF_FP_Def_Sugar.co_datatype_cmd
    BNF_Util.Least_FP
    BNF_LFP.construct_lfp
    (Ctr_Sugar.default_ctr_options_cmd,
     (map (fn ((n, v), l) =>
            ( ( ( ((map (fn v => (SOME (To_binding ""), (To_string0 v, NONE))) v, To_sbinding n), NoSyn)
                , List.map (fn (n, l) => ( ( (To_binding "", To_sbinding n)
                                           , List.map (fn s => (To_binding "", of_semi__typ s)) l)
                                         , NoSyn)) l)
              , (To_binding "", To_binding "", To_binding ""))
            , [])) l)))
  |> \<^cancel>\<open>| _ => #theory top\<close>
  (\<^cancel>\<open>snd oo Old_Datatype.add_datatype_cmd\<close>
     (tap
       (fn _ => case version of Datatype_new => ()
                              | _ => warning "Type of datatype not available in this running version of Isabelle")
       ) \<^cancel>\<open>
    (map (fn ((n, v), l) =>
           ( (To_sbinding n, map (fn v => (To_string0 v, NONE)) v, NoSyn)
           , List.map (fn (n, l) => (To_sbinding n, List.map of_semi__typ l, NoSyn)) l))
         l)\<close>)

fun type_synonym top (Type_synonym ((n, v), l)) = #theory top (fn thy => let val s_bind = To_sbinding n in
  (snd o Typedecl.abbrev_global
           (s_bind, map To_string0 v, NoSyn)
           (Isabelle_Typedecl.abbrev_cmd0 (SOME s_bind) thy (of_semi__typ l))) thy end)

fun type_notation top (Type_notation (n, e)) = #local_theory top NONE NONE
  (Local_Theory.type_notation_cmd true ("", true) [(To_string0 n, Mixfix (Input.string (To_string0 e), [], 1000, Position.no_range))])

fun instantiation1 name thy = thy
  |> Class.instantiation ([ let val Term.Type (s, _) = Isabelle_Typedecl.abbrev_cmd0 NONE thy name in s end ],
                          [],
                          Syntax.read_sort (Proof_Context.init_global thy) "object")

fun instantiation2 name n_def expr =
  Specification.definition_cmd NONE [] [] ( (To_binding (To_string0 n_def ^ "_" ^ name ^ "_def"), [])
                                          , of_semi__term expr)

fun overloading1 n_c e_c = Overloading.overloading_cmd [(To_string0 n_c, of_semi__term e_c, true)]

fun overloading2 n e =
  #2 oo Specification.definition_cmd NONE [] [] ((To_sbinding n, []), of_semi__term e)

fun consts top (Consts (n, ty, symb)) = #theory top
  (Sign.add_consts_cmd [( To_sbinding n
                        , of_semi__typ ty
                        , Mixfix (Input.string ("(_) " ^ To_string0 symb), [], 1000, Position.no_range))])

fun definition top def = #local_theory' top NONE NONE
  let val (def, e) = case def of
      Definitiona e => (NONE, e)
    | Definition_where1 (name, (abbrev, prio), e) =>
        (SOME ( To_sbinding name
              , NONE
              , Mixfix (Input.string ("(1" ^ of_semi__term abbrev ^ ")"), [], To_nat prio, Position.no_range)), e)
    | Definition_where2 (name, abbrev, e) =>
        (SOME ( To_sbinding name
              , NONE
              , Mixfix (Input.string ("(" ^ of_semi__term abbrev ^ ")"), [], 1000, Position.no_range)), e) in fn ctxt => ctxt
  |> #2 oo Specification.definition_cmd def [] [] (Binding.empty_atts, of_semi__term e) end

fun lemmas top lemmas = #local_theory' top NONE NONE (fn disp => fn lthy =>
  let val (simp, s, l) =
    case lemmas of Lemmas_simp_thm (simp, s, l) =>
                     (simp, s, map (fn x => ([semi__thm_attribute_single lthy x], [])) l)
                 | Lemmas_simp_thms (s, l) =>
                     (true, s, map (fn x => (Proof_Context.get_thms lthy (To_string0 x), [])) l) in
  (#2 o Specification.theorems Thm.theoremK
    [((To_sbinding s, List.map (fn s => Attrib.check_src lthy [Token.make_string (s, Position.none)])
                               (if simp then ["simp", "code_unfold"] else [])),
      l)]
    []
    disp) lthy end)

fun lemma1 n l_spec = Specification.theorem_cmd true Thm.theoremK NONE (K I)
  Binding.empty_atts [] [] (Element.Shows [((To_sbinding n, [])
                                            ,[((String.concatWith (" \<Longrightarrow> ")
                                                  (List.map of_semi__term l_spec)), [])])])

fun lemma1' n l_spec concl = Specification.theorem_cmd true Thm.theoremK NONE (K I)
  (To_sbinding n, [])
  []
  (List.map (fn (n, (b, e)) =>
              Element.Assumes [( ( To_sbinding n
                                 , if b then [[Token.make_string ("simp", Position.none)]] else [])
                               , [(of_semi__term e, [])])])
            l_spec)
  (Element.Shows [(Binding.empty_atts,[(of_semi__term concl, [])])])

fun lemma3 l_apply = map_filter (fn META.Command_let _ => SOME []
                                  | META.Command_have _ => SOME []
                                  | META.Command_fix_let (_, _, _, l) => SOME l
                                  | _ => NONE)
                                (rev l_apply)

fun axiomatization top (Axiomatization (n, e)) = #theory top
  (#2 o Specification.axiomatization_cmd [] [] [] [((To_sbinding n, []), of_semi__term e)])

fun section n s _ =
  let fun mk s n = if n <= 0 then s else mk ("  " ^ s) (n - 1) in
    out_intensify (mk (Markup.markup Markup.keyword3 (To_string0 s)) n) ""
  end

fun text name s =
  Document_Output.document_output
    {markdown = true, markup = fn body => [XML.Elem (Markup.latex_body name, body)]}
    (NONE, Input.string (To_string0 s))

fun ml top (SMLa ml) = #generic_theory top
  (ML_Context.exec let val source = input_source ml in
                   fn () => ML_Context.eval_source (ML_Compiler.verbose true ML_Compiler.flags) source
                   end #>
    Local_Theory.propagate_ml_env)

fun setup top (Setup ml) = #theory top (Isar_Cmd.setup (input_source ml))

fun thm top (Thm thm) = #keep top (fn state =>
  let val lthy = #context_of top state in
    Print_Mode.with_modes [] (fn () => writeln
      (Pretty.string_of
        (Proof_Context.pretty_fact lthy ("", List.map (semi__thm_attribute_single lthy) thm)))) ()
  end)

fun interpretation1 n loc_n loc_param =
  Interpretation.interpretation_cmd ( [ ( (To_string0 loc_n, Position.none)
                                        , ( (To_string0 n, true)
                                          , ( if loc_param = [] then
                                                Expression.Named []
                                              else
                                                Expression.Positional (map (SOME o of_semi__term)
                                                                           loc_param)
                                            , [])))]
                                    , [])

fun hide_const top (Hide_const (fully, args)) = #theory top (fn thy =>
  fold (Sign.hide_const (not fully) o ((#1 o dest_Const) oo Proof_Context.read_const {proper = true, strict = false})
                                        (Proof_Context.init_global thy))
       (map To_string0 args)
       thy)

fun abbreviation top (Abbreviation e) = #local_theory' top NONE NONE
  (Specification.abbreviation_cmd ("", true) NONE [] (of_semi__term e))

fun code_reflect' top (Code_reflect (all_public, module_name, raw_functions)) = #theory top
  (Code_Runtime'.code_reflect_cmd all_public [] (map To_string0 raw_functions) (To_string0 module_name) NONE)

end

structure Command_Transition = struct

fun semi__theory (top : ('transitionM, 'transitionM, 'state) toplevel) = let open META open META_overload
 in (*let val f = *)fn
  Theory_datatype datatype' =>
  cons (\<^command_keyword>\<open>datatype\<close>, Cmd.datatype' top datatype')
| Theory_type_synonym type_synonym => (*Toplevel.local_theory*)
  cons (\<^command_keyword>\<open>type_synonym\<close>, Cmd.type_synonym top type_synonym)
| Theory_type_notation type_notation =>
  cons (\<^command_keyword>\<open>type_notation\<close>, Cmd.type_notation top type_notation)
| Theory_instantiation (Instantiation (n, n_def, expr)) => let val name = To_string0 n in fn acc => acc
  |>:: (\<^command_keyword>\<open>instantiation\<close>, #begin_main_target top true (Cmd.instantiation1 name))
  |>:: (\<^command_keyword>\<open>definition\<close>, #local_theory' top NONE NONE (#2 oo Cmd.instantiation2 name n_def expr))
  |>:: (\<^command_keyword>\<open>instance\<close>, #local_theory_to_proof top NONE NONE (Class.instantiation_instance I))
  |>:: (\<^command_keyword>\<open>..\<close>, #tr_raw top Isar_Cmd.default_proof)
  |>:: end' top end
| Theory_overloading (Overloading (n_c, e_c, n, e)) => (fn acc => acc
  |>:: (\<^command_keyword>\<open>overloading\<close>, #begin_main_target top true (Cmd.overloading1 n_c e_c))
  |>:: (\<^command_keyword>\<open>definition\<close>, #local_theory' top NONE NONE (Cmd.overloading2 n e))
  |>:: end' top)
| Theory_consts consts =>
  cons (\<^command_keyword>\<open>consts\<close>, Cmd.consts top consts)
| Theory_definition definition =>
  cons (\<^command_keyword>\<open>definition\<close>, Cmd.definition top definition)
| Theory_lemmas lemmas =>
  cons (\<^command_keyword>\<open>lemmas\<close>, Cmd.lemmas top lemmas)
| Theory_lemma (Lemma (n, l_spec, l_apply, o_by)) => (fn acc => acc
  |>:: (\<^command_keyword>\<open>lemma\<close>, #local_theory_to_proof' top NONE NONE (Cmd.lemma1 n l_spec))
  |> fold (semi__command_proof top o META.Command_apply) l_apply
  |>:: terminal_proof_dual top o_by)
| Theory_lemma (Lemma_assumes (n, l_spec, concl, l_apply, o_by)) => (fn acc => acc
  |>:: (\<^command_keyword>\<open>lemma\<close>, #local_theory_to_proof' top NONE NONE (Cmd.lemma1' n l_spec concl))
  |> fold (semi__command_proof top) l_apply
  |> (fn st => st
    |>:: terminal_proof_dual top o_by
    |> (case Cmd.lemma3 l_apply of
        [] => I
      | _ :: l =>
        let fun cons_qed m =
  cons (\<^command_keyword>\<open>qed\<close>, #tr_report_o top m (#tr_raw top (Isar_Cmd.qed m))) in fn st => st
        |> fold (fn l => fold (semi__command_state top) l o cons_qed NONE) l
        |> cons_qed NONE end)))
| Theory_axiomatization axiomatization =>
  cons (\<^command_keyword>\<open>axiomatization\<close>, Cmd.axiomatization top axiomatization)
| Theory_section (Section (n, s)) => let val n = To_nat n in fn st => st
  |>:: let val (name, pos) = case n of 0 =>
        \<^command_keyword>\<open>section\<close> | 1 =>
        \<^command_keyword>\<open>subsection\<close> | _ =>
        \<^command_keyword>\<open>subsubsection\<close>
       in
         ( (name, pos)
         , #tr_raw
             top
             (Document_Output.document_output
               {markdown = false, markup = fn body => [XML.Elem (Markup.latex_heading name, body)]}
               (NONE, Input.string (To_string0 s))))
       end
  |>:: (\<^command_keyword>\<open>print_syntax\<close>, #keep top (Cmd.section n s)) end
| Theory_text (Text s) =>
  cons let val (name, pos) =
        \<^command_keyword>\<open>text\<close>
       in ((name, pos), #tr_raw top (Cmd.text name s)) end
| Theory_text_raw (Text_raw s) =>
  cons let val (name, pos) =
        \<^command_keyword>\<open>text_raw\<close>
       in ((name, pos), #tr_raw top (Cmd.text name s)) end
| Theory_ML ml =>
  cons (\<^command_keyword>\<open>ML\<close>, Cmd.ml top ml)
| Theory_setup setup =>
  cons (\<^command_keyword>\<open>setup\<close>, Cmd.setup top setup)
| Theory_thm thm =>
  cons (\<^command_keyword>\<open>thm\<close>, Cmd.thm top thm)
| Theory_interpretation (Interpretation (n, loc_n, loc_param, o_by)) => (fn st => st
  |>:: (\<^command_keyword>\<open>interpretation\<close>, #local_theory_to_proof top NONE NONE
     (Cmd.interpretation1 n loc_n loc_param))
  |>:: terminal_proof_dual top o_by)
| Theory_hide_const hide_const =>
  cons (\<^command_keyword>\<open>hide_const\<close>, Cmd.hide_const top hide_const)
| Theory_abbreviation abbreviation =>
  cons (\<^command_keyword>\<open>abbreviation\<close>, Cmd.abbreviation top abbreviation)
| Theory_code_reflect code_reflect' =>
  cons (\<^command_keyword>\<open>code_reflect'\<close>, Cmd.code_reflect' top code_reflect')
(*in fn t => fn thy => f t thy handle ERROR s => (warning s; thy)
 end*)
end
end

structure Command_Theory = struct

fun local_terminal_proof o_by = let open META in case o_by of
   Command_done => Proof.local_done_proof
 | Command_sorry => Proof.local_skip_proof true
 | Command_by l_apply => Proof.local_terminal_proof (then_tactic l_apply, NONE)
end

fun global_terminal_proof o_by = let open META in case o_by of
   Command_done => Proof.global_done_proof
 | Command_sorry => Proof.global_skip_proof true
 | Command_by l_apply => Proof.global_terminal_proof (then_tactic l_apply, NONE)
end

fun semi__command_state' top pr = fold snd (rev (semi__command_state top pr []))
fun semi__command_proof' top pr = fold snd (rev (semi__command_proof top pr []))

fun semi__theory top = let open META open META_overload in (*let val f = *)fn
  Theory_datatype datatype' => Cmd.datatype' top datatype'
| Theory_type_synonym type_synonym => Cmd.type_synonym top type_synonym
| Theory_type_notation type_notation => Cmd.type_notation top type_notation
| Theory_instantiation (Instantiation (n, n_def, expr)) => #theory top (fn thy => let val name = To_string0 n in thy
  |> Cmd.instantiation1 name
  |> (fn thy => let val ((_, (_, ty)), thy) = Cmd.instantiation2 name n_def expr false thy in ([ty], thy) end)
  |-> Class.prove_instantiation_exit_result (map o Morphism.thm) (fn ctxt => fn thms =>
       Class.intro_classes_tac ctxt [] THEN ALLGOALS (Proof_Context.fact_tac ctxt thms))
  |-> K I end)
| Theory_overloading (Overloading (n_c, e_c, n, e)) => #theory top (fn thy => thy
  |> Cmd.overloading1 n_c e_c
  |> Cmd.overloading2 n e false
  |> Local_Theory.exit_global)
| Theory_consts consts => Cmd.consts top consts
| Theory_definition definition => Cmd.definition top definition
| Theory_lemmas lemmas => Cmd.lemmas top lemmas
| Theory_lemma (Lemma (n, l_spec, l_apply, o_by)) => #local_theory top NONE NONE (fn lthy => lthy
  |> Cmd.lemma1 n l_spec false
  |> fold (semi__command_proof' top o META.Command_apply) l_apply
  |> global_terminal_proof o_by)
| Theory_lemma (Lemma_assumes (n, l_spec, concl, l_apply, o_by)) => #local_theory top NONE NONE (fn lthy => lthy
  |> Cmd.lemma1' n l_spec concl false
  |> fold (semi__command_proof' top) l_apply
  |> (case Cmd.lemma3 l_apply of
        [] => global_terminal_proof o_by
      | _ :: l => let val arg = (NONE, true) in fn st => st
        |> local_terminal_proof o_by
        |> fold (fn l => fold (semi__command_state' top) l o Proof.local_qed arg) l
        |> Proof.global_qed arg end))
| Theory_axiomatization axiomatization => Cmd.axiomatization top axiomatization
| Theory_section (Section (n, s)) => #keep top (Cmd.section (To_nat n) s)
| Theory_text _ => #keep top (K ())
| Theory_text_raw _ => #keep top (K ())
| Theory_ML ml => Cmd.ml top ml
| Theory_setup setup => Cmd.setup top setup
| Theory_thm thm => Cmd.thm top thm
| Theory_interpretation (Interpretation (n, loc_n, loc_param, o_by)) => #local_theory top NONE NONE (fn lthy => lthy
  |> Cmd.interpretation1 n loc_n loc_param
  |> global_terminal_proof o_by)
| Theory_hide_const hide_const => Cmd.hide_const top hide_const
| Theory_abbreviation abbreviation => Cmd.abbreviation top abbreviation
| Theory_code_reflect code_reflect' => Cmd.code_reflect' top code_reflect'
(*in fn t => fn thy => f t thy handle ERROR s => (warning s; thy)
 end*)
end
end

end

structure Bind_META = struct open Bind_Isabelle

type T = META.all_meta list

structure Meta_Cmd_Data = Theory_Data
(
  open META
  type T = T
  val empty = []
  val merge = #2
)

fun ML_context_exec pos ants =
  ML_Context.exec (fn () =>
    ML_Context.eval (ML_Compiler.verbose false ML_Compiler.flags)
                    pos
                    ants) #>
        Local_Theory.propagate_ml_env

fun meta_command0 s_put f_get constraint source =
  let val name = "meta_command"
  in
  Context.Theory 
  #> ML_context_exec (Input.pos_of source)
       (ML_Lex.read "let open META val ML = META.SML val "
        @ ML_Lex.read_range (Input.range_of source) name
        @ ML_Lex.read (" : " ^ constraint ^ " = ")
        @ ML_Lex.read_source source
        @ ML_Lex.read (" in Context.>> (Context.map_theory (" ^ s_put ^ " " ^ name ^ ")) end"))
  #> Context.map_theory_result (fn thy => (f_get thy, thy))
  #> fst
  end

val meta_command = meta_command0 "Bind_META.Meta_Cmd_Data.put" Meta_Cmd_Data.get "Bind_META.T"

local
  open META
  open META_overload
  open Library

  fun semi__locale data thy = thy
           |> (   Expression.add_locale_cmd
                    (To_sbinding (META.holThyLocale_name data))
                    Binding.empty
                    []
                    ([], [])
                    (List.concat
                      (map
                        (fn (fixes, assumes) => List.concat
                          [ map (fn (e,ty) => Element.Fixes [( To_binding (of_semi__term e)
                                                             , SOME (of_semi__typ ty)
                                                             , NoSyn)]) fixes
                          , case assumes of NONE => []
                                          | SOME (n, e) => [Element.Assumes [( (To_sbinding n, [])
                                                                             , [(of_semi__term e, [])])]]])
                        (META.holThyLocale_header data)))
               #> #2)

  fun semi__aux thy =
    map2_ctxt_term
      (fn T_pure x => T_pure x
        | e =>
          let fun aux e = case e of
            T_to_be_parsed (s, _) => SOME let val t = Syntax.read_term (get_thy \<^here> Proof_Context.init_global thy)
                                                                       (To_string0 s) in
                                          (t, s, Term.add_frees t [])
                                          end
          | T_lambda (a, e) =>
            Option.map
              (fn (e, s, l_free) =>
               let val a0 = To_string0 a
                   val (t, l_free) = case List.partition (fn (x, _) => x = a0) l_free of
                                       ([], l_free) => (Term.TFree ("'a", ["HOL.type"]), l_free)
                                     | ([(_, t)], l_free) => (t, l_free) in
               (lambda ( Term.Free (a0, t)) e
                       , META.String_concatWith (From.string "", [From.string "(% ", a, From.string ". ", s, From.string ")"])
                       , l_free)
               end)
              (aux e)
          | _ => NONE in
          case aux e of
            NONE => error "nested pure expression not expected"
          | SOME (e, s, _) => META.T_pure (From.Pure.term e, SOME s)
          end)
in

fun all_meta_tr aux top thy_o = fn
  META_semi_theories theo => apsnd
    (case theo of
       Theories_one theo => Command_Transition.semi__theory top theo
     | Theories_locale (data, l) => fn acc => acc
       |>:: (\<^command_keyword>\<open>locale\<close>, #begin_main_target top true (semi__locale data))
       |> fold (fold (Command_Transition.semi__theory top)) l
       |>:: end' top)
| META_boot_generation_syntax _ => I
| META_boot_setup_env _ => I
| META_all_meta_embedding (META_generic (OclGeneric source)) => 
    (fn (env, tr) =>
      all_meta_trs
        aux
        top
        thy_o
        (get_thy \<^here>
                 (fn thy =>
                   get_thy \<^here>
                           (meta_command (Input.string (To_string0 source)))
                           (if forall (fn ((key, _), _) =>
                                        Keyword.is_vacuous (Thy_Header.get_keywords thy) key)
                                      tr
                            then SOME thy else NONE))
                 thy_o)
          (env, tr))
| META_all_meta_embedding meta => aux (semi__aux NONE meta)

and all_meta_trs aux = fold oo all_meta_tr aux

fun all_meta_thy aux top_theory top_local_theory = fn
  META_semi_theories theo => apsnd
    (case theo of
       Theories_one theo => Command_Theory.semi__theory top_theory theo
     | Theories_locale (data, l) => (*Toplevel.begin_main_target*) fn thy => thy
       |> semi__locale data
       |> fold (fold (Command_Theory.semi__theory top_local_theory)) l
       |> Local_Theory.exit_global)
| META_boot_generation_syntax _ => I
| META_boot_setup_env _ => I
| META_all_meta_embedding (META_generic (OclGeneric source)) =>
    (fn (env, thy) =>
      all_meta_thys aux top_theory top_local_theory (meta_command (Input.string (To_string0 source)) thy) (env, thy))
| META_all_meta_embedding meta => fn (env, thy) => aux (semi__aux (SOME thy) meta) (env, thy)

and all_meta_thys aux = fold oo all_meta_thy aux

end
end
\<close>

subsection\<open>Directives of Compilation for Target Languages\<close>

ML\<open>
structure Deep = struct

fun absolute_path thy filename =
  Path.implode (Path.append (Resources.master_directory thy) (Path.explode filename))

val list_all_eq = fn [] => true
                   | x0 :: xs => List.all (fn x1 => x0 = x1) xs

end
\<close>

subsection\<open>Saving the History of Meta Commands\<close>

ML\<open>
fun p_gen f g =  f "[" "]" g
              (*|| f "{" "}" g*)
              || f "(" ")" g
fun paren f = p_gen (fn s1 => fn s2 => fn f => Parse.$$$ s1 |-- f --| Parse.$$$ s2) f
fun parse_l f = Parse.$$$ "[" |-- Parse.!!! (Parse.list f --| Parse.$$$ "]")
fun parse_l_with f = Parse.$$$ "[" |-- Scan.optional (Parse.binding --| \<^keyword>\<open>with_only\<close> >> SOME) NONE
                     -- Parse.!!! (Parse.list f --| Parse.$$$ "]")
fun parse_l' f = Parse.$$$ "[" |-- Parse.list f --| Parse.$$$ "]"
fun parse_l1' f = Parse.$$$ "[" |-- Parse.list1 f --| Parse.$$$ "]"
fun annot_ty f = Parse.$$$ "(" |-- f --| Parse.$$$ "::" -- Parse.binding --| Parse.$$$ ")"
\<close>

ML\<open>
structure Generation_mode = struct

type internal_deep =
  { output_header_thy : (string * (string list (* imports *) * string (* import optional (bootstrap) *))) option
  , seri_args : ((bstring (* compiler *) * bstring (* main module *) ) * Token.T list) list
  , filename_thy : bstring option
  , tmp_export_code : Path.T (* dir *)
  , skip_exportation : bool (* true: skip preview of code exportation *) }

datatype ('a, 'b, 'c) generation_mode0 = Gen_deep of 'a | Gen_shallow of 'b | Gen_syntax_print of 'c

type ('compiler_env_config_ext, 'a) generation_mode =
  { deep : ('compiler_env_config_ext * internal_deep) list
  , shallow : ('compiler_env_config_ext * 'a (* theory init *)) list
  , syntax_print : int option list }

fun mapM_syntax_print f (mode : ('compiler_env_config_ext, 'a) generation_mode) tr = tr
  |> f (#syntax_print mode)
  |> apfst (fn syntax_print => { syntax_print = syntax_print
                               , deep = #deep mode
                               , shallow = #shallow mode })

fun mapM_shallow f (mode : ('compiler_env_config_ext, 'a) generation_mode) tr = tr
  |> f (#shallow mode)
  |> apfst (fn shallow => { syntax_print = #syntax_print mode
                          , deep = #deep mode
                          , shallow = shallow })

fun mapM_deep f (mode : ('compiler_env_config_ext, 'a) generation_mode) tr = tr
  |> f (#deep mode)
  |> apfst (fn deep => { syntax_print = #syntax_print mode
                       , deep = deep
                       , shallow = #shallow mode })

type T = (unit META.compiler_env_config_ext, theory) generation_mode

structure Data_gen = Theory_Data
(
  type T = T
  val empty = {deep = [], shallow = [], syntax_print = [NONE]}
  fun merge (e1, e2) = { deep = #deep e1 @ #deep e2
                       , shallow = #shallow e1 @ #shallow e2
                       , syntax_print = #syntax_print e1 @ #syntax_print e2 }
)

val code_expr_argsP = Scan.optional (\<^keyword>\<open>(\<close> |-- Parse.args --| \<^keyword>\<open>)\<close>) []

val parse_scheme =
  \<^keyword>\<open>design\<close> >> K META.Gen_only_design || \<^keyword>\<open>analysis\<close> >> K META.Gen_only_analysis

val parse_sorry_mode =
  Scan.optional (  \<^keyword>\<open>SORRY\<close> >> K (SOME META.Gen_sorry)
                || \<^keyword>\<open>no_dirty\<close> >> K (SOME META.Gen_no_dirty)) NONE

val parse_deep =
     Scan.optional (\<^keyword>\<open>skip_export\<close> >> K true) false
  -- Scan.optional (((Parse.$$$ "(" -- \<^keyword>\<open>THEORY\<close>) |-- Parse.name -- ((Parse.$$$ ")"
                   -- Parse.$$$ "(" -- \<^keyword>\<open>IMPORTS\<close>) |-- parse_l' Parse.name -- Parse.name)
                   --| Parse.$$$ ")") >> SOME) NONE
  -- Scan.optional (\<^keyword>\<open>SECTION\<close> >> K true) false
  -- parse_sorry_mode
  -- (* code_expr_inP *) parse_l1' (\<^keyword>\<open>in\<close> |-- ((\<^keyword>\<open>self\<close> || Parse.name)
        -- Scan.optional (\<^keyword>\<open>module_name\<close> |-- Parse.name) ""
        -- code_expr_argsP))
  -- Scan.optional
       ((Parse.$$$ "(" -- \<^keyword>\<open>output_directory\<close>) |-- Parse.name --| Parse.$$$ ")" >> SOME)
       NONE

val parse_semantics =
  let val z = 0 in
      Scan.optional
        (paren (\<^keyword>\<open>generation_semantics\<close>
               |-- paren (parse_scheme
                          -- Scan.optional ((Parse.$$$ "," -- \<^keyword>\<open>oid_start\<close>) |-- Parse.nat)
                                           z)))
              (META.Gen_default, z)
  end

val mode =
  let fun mk_env output_disable_thy output_header_thy oid_start design_analysis sorry_mode ctxt =
    META.compiler_env_config_empty
    output_disable_thy
    (From.option (From.pair From.string (From.pair (From.list From.string) From.string))
                 output_header_thy)
    (META.oidInit (From.internal_oid oid_start))
    design_analysis
    (sorry_mode, Config.get ctxt quick_and_dirty) in

     \<^keyword>\<open>deep\<close> |-- parse_semantics -- parse_deep >>
     (fn ( (design_analysis, oid_start)
         , ( ((((skip_exportation, output_header_thy), output_disable_thy), sorry_mode), seri_args)
           , filename_thy)) =>
         Gen_deep ( mk_env (not output_disable_thy)
                           output_header_thy
                           oid_start
                           design_analysis
                           sorry_mode
                  , { output_header_thy = output_header_thy
                    , seri_args = seri_args
                    , filename_thy = filename_thy
                    , tmp_export_code = Isabelle_System.create_tmp_path "deep_export_code" ""
                    , skip_exportation = skip_exportation }))
  || \<^keyword>\<open>shallow\<close> |-- parse_semantics -- parse_sorry_mode >>
     (fn ((design_analysis, oid_start), sorry_mode) =>
       Gen_shallow (mk_env true
                           NONE
                           oid_start
                           design_analysis
                           sorry_mode))
  || (\<^keyword>\<open>syntax_print\<close> |-- Scan.optional (Parse.number >> SOME) NONE) >>
     (fn n => Gen_syntax_print (case n of NONE => NONE | SOME n => Int.fromString n))
  end

fun f_command l_mode =
  Toplevel'.setup_theory
    (META.mapM
      (fn Gen_shallow env =>
           pair (fn thy => Gen_shallow (env (Proof_Context.init_global thy), thy))
                o cons (Toplevel'.read_write_keep (Toplevel'.Load_previous, Toplevel'.Store_backup))
        | Gen_syntax_print n => pair (K (Gen_syntax_print n))
        | Gen_deep (env, i_deep) =>
           pair (fn thy => Gen_deep (env (Proof_Context.init_global thy), i_deep))
                o cons
            (\<^command_keyword>\<open>export_code\<close>, Toplevel'.keep_theory (fn _ =>
              let val seri_args' =
                      List.filter (fn (("self", _), _) => false | _ => true) (#seri_args i_deep)
              in case seri_args' of [] => () | _ =>
                      warning ("Ignoring the exportation to target languages")
              end)))
      l_mode
      [])
    (fn l_mode => fn thy =>
      let val l_mode = map (fn f => f thy) l_mode
      in Data_gen.put { deep = map_filter (fn Gen_deep x => SOME x | _ => NONE) l_mode
                      , shallow = map_filter (fn Gen_shallow x => SOME x | _ => NONE) l_mode
                      , syntax_print = map_filter (fn Gen_syntax_print x => SOME x | _ => NONE) l_mode } thy end)

fun update_compiler_config f =
  Data_gen.map
    (fn mode => { deep = map (apfst (META.compiler_env_config_update f)) (#deep mode)
                , shallow = map (apfst (META.compiler_env_config_update f)) (#shallow mode)
                , syntax_print = #syntax_print mode })

fun meta_command0 s_put f_get f_get0 constraint source =
  let val name = "meta_command'"
  in
  Context.Theory 
  #> Bind_META.ML_context_exec (Input.pos_of source)
       (ML_Lex.read "let open META val ML = META.SML val "
        @ ML_Lex.read_range (Input.range_of source) name
        @ ML_Lex.read (" : " ^ constraint ^ " = ")
        @ ML_Lex.read_source source
        @ ML_Lex.read (" in Context.>> (Context.map_theory (fn thy => " ^ s_put ^ " (" ^ name ^ " (" ^ f_get0 ^ " thy)) thy)) end"))
  #> Context.map_theory_result (fn thy => (f_get thy, thy))
  #> fst
  end

val meta_command = meta_command0 "Bind_META.Meta_Cmd_Data.put"
                                 Bind_META.Meta_Cmd_Data.get
                                 "Generation_mode.Data_gen.get"
                                 "Generation_mode.T -> Bind_META.T"
end
\<close>

subsection\<open>Factoring All Meta Commands Together\<close>

setup\<open>ML_Antiquotation.inline \<^binding>\<open>mk_string\<close> (Scan.succeed
"(fn ctxt => fn x => ML_Pretty.string_of_polyml (ML_system_pretty (x, FixedInt.fromInt (Config.get ctxt ML_Print_Depth.print_depth))))")
\<close>

ML\<open>

local
  val partition_self = List.partition (fn ((s,_),_) => s = "self")
in

fun exec_deep0 {output_header_thy, seri_args, filename_thy, ...} (env, l_obj) =
let open Generation_mode
    val (seri_args0, _) = partition_self seri_args
 in
  fn thy0 =>
  let
    val env = META.compiler_env_config_more_map
                         (fn () => (l_obj, From.option
                                             From.string
                                             (Option.map (Deep.absolute_path thy0) filename_thy)))
                         env
    val l = []
  in
    case seri_args0 of [] => l
    | _ => ([], case (output_header_thy, filename_thy) of
                  (SOME _, SOME _) => let val _ = META.write_file env in "" end
                | _ => String.concat (map (fn s => s ^ "\n") (snd (META.write_file0 env)))
                       (* TODO: further optimize "string" as "string list" *))
           :: l
  end
  |> (fn l => let val (l_warn, l) = (map fst l, map snd l) in
      if Deep.list_all_eq l then
        (List.concat l_warn, case l of [] => "" | x :: _ => x)
      else
        error "There is an extracted language which does not produce a similar Isabelle content as the others"
      end)
  |> (fn (l_warn, s) =>
       let val () = writeln
         (case (output_header_thy, filename_thy) of
            (SOME _, SOME _) => s
          | _ => String.concat (map ( (fn s => s ^ "\n")
                                    o Active.sendback_markup_command
                                    o trim_line)
                                    (String.tokens (fn c => Char.ord c = META.integer_escape) s)))
       in List.app (fn (out, err) => ( writeln (Markup.markup Markup.keyword2 err)
                                     ; case trim_line out of "" => ()
                                       | out => writeln (Markup.markup Markup.keyword1 out)))
                   l_warn end)
end

fun exec_deep i_deep e =
  let val (seri_args0, seri_args) = partition_self (#seri_args i_deep)
  in cons
      ( case (seri_args0, seri_args) of ([_], []) => \<^command_keyword>\<open>print_syntax\<close>
                                      | _ => \<^command_keyword>\<open>export_code\<close>
      , Toplevel'.keep_theory (exec_deep0 i_deep e))
  end
end

local

fun fold_thy_shallow f =
  META.fold_thy_shallow
    (fn f => f () handle ERROR e =>
      ( warning "Shallow Backtracking: (true) Isabelle declarations occurring among the META-simulated ones are ignored (if any)"
        (* TODO automatically determine if there is such Isabelle declarations,
                for raising earlier a specific error message *)
      ; error e))
    f

fun disp_time toplevel_keep_output =
  let
    val tps = Timing.start ()
    val disp_time = fn NONE => I | SOME msg =>
      toplevel_keep_output tps Markup.antiquote
        let val msg = To_string0 msg
        in " " ^ Pretty.string_of
             (Pretty.mark (Name_Space.markup (Proof_Context.const_space \<^context>) msg)
                          (Pretty.str msg)) end
  in (tps, disp_time) end

fun thy_deep exec_deep exec_info l_obj =
  Generation_mode.mapM_deep
    (META.mapM (fn (env, i_deep) =>
      pair (META.fold_thy_deep l_obj env, i_deep)
           o (if #skip_exportation i_deep then
                I
              else
                let fun exec l_obj =
                  exec_deep { output_header_thy = #output_header_thy i_deep
                            , seri_args = #seri_args i_deep
                            , filename_thy = NONE
                            , tmp_export_code = #tmp_export_code i_deep
                            , skip_exportation = #skip_exportation i_deep }
                            ( META.d_output_header_thy_update (K NONE) env, l_obj)
                in
                  case l_obj of
                    META.Fold_meta obj => exec [obj]
                  | META.Fold_custom l_obj =>
                      let val l_obj' = map_filter (fn META.META_all_meta_embedding x => SOME x
                                                    | _ => NONE)
                                                  l_obj
                      in if length l_obj' = length l_obj
                         then exec l_obj'
                         else
                           exec_info
                             (fn _ =>
                               app ( writeln
                                   o Active.sendback_markup_command
                                   o META.print META.of_all_meta (META.d_output_header_thy_update (K NONE) env))
                                   l_obj)
                      end
                end)))

fun report m f = (Method.report m; f)
fun report_o o' f = (Option.map Method.report o'; f)

fun thy_shallow l_obj get_all_meta_embed =
  Generation_mode.mapM_shallow
    (fn l_shallow => fn thy => META.mapM
      (fn (env, thy0) => fn (thy, l_obj) =>
        let val (_, disp_time) = disp_time (tap o K ooo out_intensify')
            fun aux x =
              fold_thy_shallow
                (K o K thy0)
                (fn msg =>
                  let val () = disp_time msg ()
                      fun in_self f =
                        \<comment> \<open>Note: This function is not equivalent to \<^ML>\<open>Local_Theory.begin_nested\<close> \<^ML>\<open>Local_Theory.end_nested\<close>.\<close>
                        Local_Theory.new_group
                        #> f
                        #> Local_Theory.reset_group
                        #> (fn lthy =>
                            #1 (Target_Context.switch_named_cmd NONE (Context.Proof lthy)) lthy |> Context.the_proof)
                      fun not_used p _ = error ("not used " ^ Position.here p)
                      val context_of = I
                      fun proof' f = f true
                      fun proofs f s = s |> f |> Seq.the_result ""
                      val proof = I
                      val dual = #seq in
                  Bind_META.all_meta_thys (aux o META.Fold_meta)

                                          { (* specialized part *)
                                            theory = I
                                          , local_theory = K o K Named_Target.theory_map
                                          , local_theory' = K o K (fn f => Named_Target.theory_map (f false))
                                          , keep = fn f => Named_Target.theory_map (fn lthy => (f lthy ; lthy))
                                          , generic_theory = Context.theory_map
                                            (* generic part *)
                                          , context_of = context_of, dual = dual
                                          , proof' = proof', proofs = proofs, proof = proof
                                          , tr_report = report, tr_report_o = report_o
                                          , pr_report = report, pr_report_o = report_o
                                            (* irrelevant part *)
                                          , begin_main_target = K o not_used \<^here>
                                          , local_theory_to_proof' = K o K not_used \<^here>
                                          , local_theory_to_proof = K o K not_used \<^here>
                                          , tr_raw = not_used \<^here> }

                                          { (* specialized part *)
                                            theory = Local_Theory.background_theory
                                          , local_theory = K o K in_self
                                          , local_theory' = K o K (fn f => in_self (f false))
                                          , keep = fn f => in_self (fn lthy => (f lthy ; lthy))
                                          , generic_theory = Context.proof_map
                                            (* generic part *)
                                          , context_of = context_of, dual = dual
                                          , proof' = proof', proofs = proofs, proof = proof
                                          , tr_report = report, tr_report_o = report_o
                                          , pr_report = report, pr_report_o = report_o
                                            (* irrelevant part *)
                                          , begin_main_target = K o not_used \<^here>
                                          , local_theory_to_proof' = K o K not_used \<^here>
                                          , local_theory_to_proof = K o K not_used \<^here>
                                          , tr_raw = not_used \<^here> }
                  end)
                x
            val (env, thy) =
              let
                fun disp_time f x =
                let val (s, r) = Timing.timing f x
                    val () = out_intensify (Timing.message s |> Markup.markup Markup.operator) "" in
                  r
                end
              in disp_time (fn x => aux x (env, thy)) (l_obj ()) end
        in ((env, thy0), (thy, fn _ => get_all_meta_embed (SOME thy))) end)
      l_shallow
      (thy, case l_obj of SOME f => f | NONE => fn _ => get_all_meta_embed (SOME thy))
      |> META.map_prod I fst)

fun thy_switch \<^cancel>\<open>pos1 pos2\<close> f mode tr =
  ( ( mode
    , \<^cancel>\<open>Toplevel'.keep
        (fn _ => Output.information ( "Theory required while transitions were being built"
                                    ^ Position.here pos1
                                    ^ ": Commands will not be concurrently considered. "
                                    ^ Markup.markup
                                        (Markup.properties (Position.properties_of pos2) Markup.position)
                                        "(Handled here\092<^here>)"))\<close> tr)
  , f #~> Generation_mode.Data_gen.put)

in

fun outer_syntax_commands'''' is_safe mk_string get_all_meta_embed name thy _ =
  (* WARNING: Whenever there would be errors raised by functions taking "thy" as input,
              they will not be shown.
              So the use of this "thy" can be considered as safe, as long as errors do not happen. *)
  let
    open Generation_mode
    val get_all_m = get_all_meta_embed name
    val m_tr = (Data_gen.get thy, [])
      |-> mapM_syntax_print (META.mapM (fn n =>
            pair n
                 o cons (\<^command_keyword>\<open>print_syntax\<close>,
                         Toplevel'.keep_theory (fn thy =>
                           writeln (mk_string
                                     (Proof_Context.init_global
                                       (case n of NONE => thy
                                                | SOME n => Config.put_global ML_Print_Depth.print_depth n thy))
                                     name)))))
  in \<^cancel>\<open>let
       val thy_o = is_safe thy
       val l_obj = get_all_m thy_o
                   (* In principle, it is fine if (SOME thy) is provided to
                      get_all_m. However, because certain types of errors are most of the
                      time happening whenever certain specific operations depending on thy
                      are explicitly performed, and because get_all_m was intentionally set
                      to not interactively manage such errors, then these errors (whenever
                      they are happening) could possibly not appear in the output
                      window. Although the computation would be in any case interrupted as
                      usual (but with only minimal debugging information, such as a simple
                      red underlining color).
                      
                      Generally, whenever get_all_m is called during the evaluating commands
                      coming from generated files (which is not the case here, but will be
                      later), this restriction can normally be removed (i.e., by writing
                      (SOME thy)), as for the case of generated files, we are taking the
                      assumption that errors (if they are happening) are as hard to detect
                      as if an error was raised somewhere else by the generator itself.
                      Another assumption nevertheless related with the generator is that it
                      is supposed to explicitly not raise errors, however here this
                      get_all_m is not situated below a generating part. This is why we are
                      tempted to mostly give NONE to get_all_m, unless the calling command
                      is explicitly taking the responsibility of a potential failure. *)
       val m_tr = m_tr
                  |-> thy_deep exec_deep Toplevel'.keep l_obj
     in ( m_tr
          |-> mapM_shallow (META.mapM (fn (env, thy_init) => fn acc =>
                let val (tps, disp_time) = disp_time Toplevel'.keep_output
                    fun aux thy_o =
                      fold_thy_shallow
                        (K (cons (Toplevel'.read_write_keep (Toplevel.Load_backup, Toplevel.Store_default))))
                        (fn msg => fn l =>
                          apsnd (disp_time msg)
                          #> Bind_META.all_meta_trs (aux thy_o o META.Fold_meta)
                                                    { context_of = Toplevel.context_of
                                                    , keep = Toplevel.keep
                                                    , generic_theory = Toplevel.generic_theory
                                                    , theory = Toplevel.theory
                                                    , begin_main_target = Toplevel.begin_main_target
                                                    , local_theory' = Toplevel.local_theory'
                                                    , local_theory = Toplevel.local_theory
                                                    , local_theory_to_proof' = Toplevel.local_theory_to_proof'
                                                    , local_theory_to_proof = Toplevel.local_theory_to_proof
                                                    , proof' = Toplevel.proof'
                                                    , proofs = Toplevel.proofs
                                                    , proof = Toplevel.proof
                                                      (* *)
                                                    , dual = #par, tr_raw = I
                                                    , tr_report = report, tr_report_o = report_o
                                                    , pr_report = report, pr_report_o = report_o }
                                                    thy_o
                                                    l)
                in aux thy_o l_obj (env, acc)
                   |> META.map_prod
                        (fn env => (env, thy_init))
                        (Toplevel'.keep_output tps Markup.operator "") end))
        , Data_gen.put)
        handle THY_REQUIRED pos =>
          m_tr |-> thy_switch pos \<^here> (thy_shallow NONE get_all_m)
     end
     handle THY_REQUIRED pos =>
       \<close>m_tr |-> thy_switch \<^cancel>\<open>pos \<^here>\<close> (fn mode => fn thy => 
                                        let val l_obj = get_all_m (SOME thy) in
                                          (thy_deep (tap oo exec_deep0) tap l_obj
                                             #~> thy_shallow (SOME (K l_obj)) get_all_m) mode thy
                                        end)
  end
  |> uncurry Toplevel'.setup_theory

fun outer_syntax_commands''' is_safe mk_string cmd_spec cmd_descr parser get_all_meta_embed =
  Outer_Syntax'.command cmd_spec cmd_descr
    (parser >> outer_syntax_commands'''' is_safe mk_string get_all_meta_embed)
end

fun outer_syntax_commands'' mk_string = outer_syntax_commands''' (K NONE) mk_string

fun outer_syntax_commands' mk_string cmd_spec cmd_descr parser get_all_meta_embed =
  outer_syntax_commands'' mk_string cmd_spec cmd_descr parser (META.Fold_meta oo get_all_meta_embed)

fun outer_syntax_commands'2 mk_string cmd_spec cmd_descr parser get_all_meta_embed =
  outer_syntax_commands''' SOME mk_string cmd_spec cmd_descr parser (META.Fold_meta oo get_all_meta_embed)
\<close>

subsection\<open>Parameterizing the Semantics of Embedded Languages\<close>

ML\<open>
val () = let open Generation_mode in
  Outer_Syntax'.command \<^command_keyword>\<open>generation_syntax\<close> "set the generating list"
    ((   mode >> (fn x => SOME [x])
      || parse_l' mode >> SOME
      || \<^keyword>\<open>deep\<close> -- \<^keyword>\<open>flush_all\<close> >> K NONE) >>
    (fn SOME x => K (K (f_command x))
      | NONE => fn thy => fn _ => []
          |> fold (fn (env, i_deep) => exec_deep i_deep (META.compiler_env_config_reset_all env))
                  (#deep (Data_gen.get thy))
          |> (fn [] => Toplevel'.keep (fn _ => warning "Nothing performed.") []
               | l => l)))
end
\<close>

subsection\<open>Setup of Meta Commands for a Generic Usage: @{command meta_command}, @{command meta_command'}\<close>

ML\<open>
local
  fun outer_syntax_commands'''2 command_keyword meta_command =
    outer_syntax_commands''' SOME \<^mk_string> command_keyword ""
      Parse.ML_source
      (fn source =>
        get_thy \<^here> (meta_command source #> META.Fold_custom))
in
val () = outer_syntax_commands'''2 \<^command_keyword>\<open>meta_command\<close> Bind_META.meta_command
val () = outer_syntax_commands'''2 \<^command_keyword>\<open>meta_command'\<close> Generation_mode.meta_command
end
\<close>

section \<open>Setup of \<^theory_text>\<open>C\<close> commands\<close>

ML \<comment> \<open>\<^theory>\<open>Isabelle_C.C_Command\<close>\<close> \<open>
val clean_C99 = Attrib.setup_config_bool @{binding Clean} (K false)

val _ =
  Theory.setup
    (Context.theory_map
      (C_Module.Data_Accept.put
        (fn ast => fn env_lang => fn context =>
          if Config.get (Context.proof_of context) clean_C99 then
            let val l_meta = Clean_Core.compile (ast |> C_Grammar_Rule.get_CTranslUnit |> the) env_lang (Context.theory_name (Context.theory_of context), length (C_Module.Data_In_Source.get context))
            in Context.map_theory
                (Outer_Syntax'.command'
                  (outer_syntax_commands''''
                    SOME
                    \<^mk_string>
                    (K (get_thy \<^here> (K l_meta #> META.Fold_custom)))
                    ()))
                context
            end
          else
            context)))
\<close>

end
